Require Import Bool.
Require Import Arith.

(* Apply also works with conditional hypotheses: *)

Theorem silly2 : forall(n m o p : nat),
     n = m ->
     (forall (q r : nat), q = r -> (q,o) = (r,p)) ->     (n,o) = (m,p).
Proof.
  intros n m o p eq1 eq2.
  apply eq2. apply eq1. Qed.

(* Through unification, Coq picks appropriate values for 
   the universal variables *)

Theorem silly2a : forall (n m : nat),
     (n,n) = (m,m) ->
     (forall (q r : nat), (q,q) = (r,r) -> q = r) ->
     n = m.
Proof.
  intros n m eq1 eq2.
  apply eq2. apply eq1. Qed.

(* The goal must match the hypothesis exactly for apply to work: *)

Theorem silly3_firsttry : forall (n : nat),
     true = beq_nat n 5 ->
     beq_nat (S (S n)) 7 = true.
Proof.
  intros n H.
  simpl.
  (* Here we cannot use apply directly *)
Abort.

(* In this case we can use the symmetry tactic, 
   which switches the left and right sides of an equality in the goal. *)

Theorem silly3 : forall (n : nat),
     true = beq_nat n 5 ->
     beq_nat (S (S n)) 7 = true.
Proof.
  intros n H.
  symmetry.
  apply H. Qed.

(* The apply ... with ... Tactic
   Examples like this one use two rewrites to exploit 
  the transitivity of equality. We can abstract this into its own lemma. *)

Example trans_eq_example : forall(a b c d e f : nat),
     (a,b) = (c,d) ->
     (c,d) = (e,f) ->
     (a,b) = (e,f).
Proof.
  intros a b c d e f eq1 eq2.
  rewrite -> eq1. rewrite -> eq2. reflexivity. Qed.

Theorem trans_eq : forall (X:Type) (n m o : X),
  n = m -> m = o -> n = o.
Proof.
  intros X n m o eq1 eq2. rewrite -> eq1. rewrite -> eq2.
  reflexivity. Qed.

(* But applying this lemma requires a slight variation on apply: *)

Example trans_eq_example' : forall(a b c d e f : nat),
     (a,b) = (c,d) ->
     (c,d) = (e,f) ->
     (a,b) = (e,f).
Proof.
  intros a b c d e f eq1 eq2.
  (* Doing apply trans_eq doesn't work!  But... *)
  apply trans_eq with (m:=(c,d)). apply eq1. apply eq2. Qed.

(* In practice, often Coq can figure out the correct variable and 
   we can omit the "m :=" part. *)


(* The inversion tactic
In Coq, the constructors of inductive types are injective and disjoint.
E.g., for nat...

    the only way we can have S n = S m is if n = m (that is, S is injective)
    O is not equal to S n for any n (that is, O and S are disjoint)

Coq provides a tactic called inversion that allows us to exploit these principles in proofs.
The inversion tactic is used like this. Suppose H is a hypothesis in the context (or a previously proven lemma) of the form
      c a1 a2 ... an = d b1 b2 ... bm
for some constructors c and d and arguments a1 ... an and b1 ... bm. Then inversion H instructs Coq to "invert" this equality to extract the information it contains about these terms:

    If c and d are the same constructor, then we know, by the injectivity of this constructor, that a1 = b1, a2 = b2, etc.; inversion H adds these facts to the context, and tries to use them to rewrite the goal.
    If c and d are different constructors, then the hypothesis H is contradictory. That is, a false assumption has crept into the context, and this means that any goal whatsoever is provable! In this case, inversion H marks the current goal as completed and pops it off the goal stack.
*)


Theorem silly4 : forall (n m : nat),
     (n,n) = (m,m) ->
     n = m.
Proof.
  intros n o eq. inversion eq. reflexivity. Qed.

Theorem silly5 : forall (n m o : nat),
     (n,m) = (o,o) ->
     (n,n) = (m,m).
Proof.
  intros n m o eq. inversion eq. reflexivity. Qed.

Theorem silly6 : forall(n : nat),
     S n = O ->
     2 + 2 = 5.
Proof.
  intros n contra. inversion contra. Qed.

Theorem silly7 : forall (n m : nat),
     false = true ->
     n = m.
Proof.
  intros n m contra. inversion contra. Qed.


Theorem beq_nat_0_l : forall n,
   beq_nat 0 n = true -> n = 0.
Proof.
  (* FILL IN HERE *) Admitted.

Theorem beq_nat_0_r : forall n,
   beq_nat n 0 = true -> n = 0.
Proof.
  (* FILL IN HERE *) Admitted.


(* Using Tactics on Hypotheses
   Tactics often come with "...in..." variants that can be used 
   on hypotheses instead of goals. *)

Theorem S_inj : forall (n m : nat) (b : bool),
     beq_nat (S n) (S m) = b ->
     beq_nat n m = b.
Proof.
  intros n m b H. simpl in H. apply H. Qed.

(* The tactic apply is a form of "backward reasoning"; it says:
  "We're trying to prove X and we know Y→X, so if we can prove Y we'll be done." 
   By contrast, the variant apply... in... is "forward reasoning"; it says 
   "We know Y and we know Y→X, so we can deduce X." *)

Theorem silly3' : forall (n : nat),
  (beq_nat n 5 = true -> beq_nat (S (S n)) 7 = true) ->
     true = beq_nat n 5 ->
     true = beq_nat (S (S n)) 7.
Proof.
  intros n eq H.
  symmetry in H. apply eq in H. symmetry in H.
  apply H. Qed.


(* Practice using "in" variants in this exercise. *)

Theorem plus_n_n_injective : forall n m,
     n + n = m + m ->
     n = m.
Proof.
  intros n. induction n as [| n'].

    (* Hint: use the plus_n_Sm lemma *)
    (* In general: SearchAbout plus gives a list of possibly useful lemmas *)
Search plus.
Check plus_n_Sm.
    (* FILL IN HERE *) Admitted.

(* Keeping a general Induction Hypothesis
   Here's a function for doubling a natural number: *)

Fixpoint double (n:nat) :=
  match n with
  | O => O
  | S n' => S (S (double n'))
  end.

(* We want to show that double is injective — i.e., 
   that it always maps different arguments to different results. 
   The way we start this proof is a little bit delicate: *)

Theorem double_injective_FAILED : forall n m,
     double n = double m ->
     n = m.
Proof.
  intros n m. induction n as [| n'].
  simpl. intros eq. destruct m as [| m']. 
    reflexivity.
    inversion eq.
  intros eq. destruct m as [| m'].
    inversion eq.
    apply f_equal.
      (* Here we are stuck.  The induction hypothesis, IHn', does
         not give us n' = m' -- there is an extra S in the
         way -- so the goal is not provable. *)
      Abort.

(* What went wrong?
   To summarize: Trying to carry out this proof by induction on n 
   when m is already in the context doesn't work because we are trying 
   to prove a relation involving every n but just a single m.
   The good proof of double_injective leaves m in the goal statement 
   at the point where the induction tactic is invoked on n: *)

Theorem double_injective : forall n m,
     double n = double m -> 
     n = m.
Proof.
  intros n. induction n as [| n'].
  simpl. intros m eq. destruct m as [| m'].
    reflexivity.
    inversion eq.
  intros m eq. destruct m as [| m'].
    inversion eq.
    apply f_equal. apply IHn'. inversion eq. reflexivity. Qed.

(* What this teaches us is that we need to be careful about 
   using induction to try to prove something too specific: 
   If we're proving a property of n and m by induction on n, 
   we may need to leave m generic.
 
The proof of this theorem (left as an exercise) has to be treated similarly: *)

Theorem beq_nat_true : forall n m,
    beq_nat n m = true -> n = m.
Proof.
  (* FILL IN HERE *) Admitted.


(* Using destruct on Compound Expressions
   The destruct tactic can be used on expressions as well as variables: *)

Definition sillyfun (n : nat) : bool :=
  if beq_nat n 3 then false
  else if beq_nat n 5 then false
  else false.

Theorem sillyfun_false : forall (n : nat),
  sillyfun n = false.
Proof.
  intros n. unfold sillyfun.
  destruct (beq_nat n 3).
    reflexivity.
    destruct (beq_nat n 5).
      reflexivity.
      reflexivity. Qed.

(* Sometimes, doing a destruct on a compound expression (a non-variable) 
   will erase information we need to complete a proof. *)

Definition sillyfun1 (n : nat) : bool :=
  if beq_nat n 3 then true
  else if beq_nat n 5 then true
  else false.

Fixpoint evenb (n:nat) : bool :=
  match n with
  | O        => true
  | S O      => false
  | S (S n') => evenb n'
  end.

Definition oddb (n:nat) : bool   :=   negb (evenb n).

Theorem sillyfun1_odd_FAILED : forall (n : nat),
     sillyfun1 n = true ->
     oddb n = true.
Proof.
  intros n eq. unfold sillyfun1 in eq.
  destruct (beq_nat n 3).
  (* stuck... *)
Abort.

(* We can use the eqn: qualifier to save that information. *)

Theorem sillyfun1_odd : forall (n : nat),
     sillyfun1 n = true ->
     oddb n = true.
Proof.
  intros n eq. unfold sillyfun1 in eq.
  destruct (beq_nat n 3) eqn:Heqe3.
    apply beq_nat_true in Heqe3.
      rewrite -> Heqe3. reflexivity.
      destruct (beq_nat n 5) eqn:Heqe5.
          apply beq_nat_true in Heqe5.
          rewrite -> Heqe5. reflexivity.
      inversion eq. Qed.
